                              /* murat demirsoy-e1061548 */

#include <iostream.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define SUCCESSFUL 1
#define FAIL       0
#define ERROR      2

#define TRUE       1
#define FALSE      0

#define EOFILE     -1

#define START_DIRECTIVE         50
#define END_DIRECTIVE           51
#define BYTE_DIRECTIVE          52
#define WORD_DIRECTIVE          53
#define RES_BYTE_WORD_DIRECTIVE 54
#define INSTRUCTION             60

#define MAX_LABEL_LENGTH       10
#define MAX_OPCODE_LENGTH      10
#define MAX_LINE_LENGTH       100	
#define MAX_TOKEN_LENGTH       15
#define MAX_FILE_NAME_LENGTH   15
#define OPCODE_TABLE_LENGTH    50
#define SYMBOL_TABLE_LENGTH    50
#define ERROR_TABLE_LENGTH    100
#define MAX_MACHINE_CODE_LENGTH 3
#define MAX_SRCFILENAME_LENGTH 15
#define MAX_PROGRAM_NAME_LENGTH 6
#define MAX_ERR_MESSAGE_SIZE  100
#define MAX_INSTRUCTION_SIZE    6
#define MAX_TEXT_RECORD_SIZE   70
#define MAX_HEADER_RECORD_SIZE 20
#define MAX_END_RECORD_SIZE    10
	

#define VALID_LABEL            10
#define INVALID_LABEL          11
#define VALID_OPCODE           20
#define INVALID_OPCODE         21
#define VALID_LABEL_SYMBOL     40
#define INVALID_LABEL_SYMBOL   41
#define VALID_NUMERIC_SYMBOL   50
#define INVALID_NUMERIC_SYMBOL 51
#define VALID_BYTE_SYMBOL      52
#define INVALID_BYTE_SYMBOL    53






struct OpCodeData {
	char Type;
	char MachineCode[MAX_MACHINE_CODE_LENGTH];
	int  MachineCodeLength;
	int  AddressFieldsNumber;
	int  DirectiveUnitLength;
};


class Assembler {
private:
	char _Pass1();
    char _Pass2();
	
	void _GetProgramName();
	int  _GetExecAddress();
	char _EndRequired;
	

public :
	char AnyError;
	char ProgramName[ MAX_PROGRAM_NAME_LENGTH ];
	int StartAddress;
	int ProgramLength;
	int LocCtr;
	unsigned int Line;
	Assembler();
	char Assemble();
};
  


class SourceFile {
protected:
	char SrcFileName[ MAX_FILE_NAME_LENGTH ];
    FILE *_SrcFile;
public:
	void setFileName(char *);
    char Open(char *mode);
	char Close();
	char Read(char *);
};


class ObjectCode: public SourceFile {
public:
	char Write(char *);
	void Print();
};





class AssemblyListing: public ObjectCode {
public:
	void Write(char *);
	
};

class Errors {
private:
	int    _ErrNo;
	char   _ErrToken[MAX_TOKEN_LENGTH];
	char * _ErrTable[ ERROR_TABLE_LENGTH ];
public:
    Errors();
   ~Errors();
    void SetErr(int errno, char *token);
    void Print();
};

template<class DataStruct> struct LinkedData{ 
	char         Name[ MAX_TOKEN_LENGTH ];
    DataStruct   Data;
	LinkedData<DataStruct> *Next;

};



class Token {
	friend void Errors::Print();
	friend void AssemblyListing::Write(char *);
protected:
	char _Token[ MAX_TOKEN_LENGTH ];
	
	char _Error(int);
    char _Numeric(char);
	char _Alpha(char);
	char _HexaDecimal(char);
	void _ToUpper();
public:
	char Read();
};


class Label : public Token {
	friend void Assembler::_GetProgramName();
private : 
	char Validate();
public :
	char Read();
	char Pass1();
};



class OpCode : public Token {
private : 
     
public :
	char Read();
	OpCodeData *Pass1();
};


class Symbol : public Token {
friend int Assembler::_GetExecAddress();
private : 
public :
	char Read(); 
	char ValidateAsLabel();
	char ValidateAsNumeric();
	int  Operand();
	int Pass2();
};


class Comment : public Token {
private : 
public :
	char Read();
};


class SourceLine {
	friend char Token::Read();
	friend char Label::Read();
	friend char Comment::Read();
	friend void Assembler::_GetProgramName();
	friend int  Assembler::_GetExecAddress();
	friend void AssemblyListing::Write(char *);

private:
    Label   _Label;
	OpCode  _OpCode;
	Symbol  _Symbol;
	Comment _Comment;

	char    _Line[ MAX_LINE_LENGTH ];
    int     _ColCtr;
    char    _CommentOrEmptyLine();
public:
	SourceLine();
    char New();
	char Pass1();
	char Pass2(char *);
};





template< class DataStruct,int TableLength > class HashTable {
protected: 
	LinkedData<DataStruct> *_Table[TableLength];
	int _Hash(char *);
public:
	HashTable();
	~HashTable();
	char        Insert(char *, DataStruct);
	DataStruct  *Search(char *);
	void        Print();
	
};

class OpCodeTable : public HashTable< struct OpCodeData, OPCODE_TABLE_LENGTH >{
	 _Insert(char *, char,char *,int,int,int);

public:
	 OpCodeTable();
   	~OpCodeTable();

	
};

class SymbolTable : public HashTable< int, SYMBOL_TABLE_LENGTH > {
public:
	void Print();
};


