#include<stdio.h>
#include<string.h>
#include<stdlib.h>

#define DEFAULT     0
#define EOFTOKEN    1
#define EOFCHAR     2
#define INTEGER     3
#define CHARACTER   4
#define HEXADECIMAL 5
#define HEXINT      50
#define LABEL       6
#define ERROR       7
#define COMMENT     8
#define COMMA       9
#define OPERATOR1   10
#define OPERATOR2   11
#define BYTE        12
#define WORD        13
#define RESB        14
#define RESW        15
#define START       16
#define END         17
#define WHITESPACE  20

class Token
{
 char *lexeme;
 int type;
 int linenum;
 unsigned char opcode;
 public :
  Token();
  Token(char * name);
  Token(char * name,int type);
  ~Token();
  void setName(char *name);
  void setType(int type);
  void setLineNumber(int line);
  void setOpcode(unsigned char);
  char * getName(void);
  int  getType(void);
  int getLineNumber(void);
  unsigned char getOpcode(void);
};

Token::Token()
{
 lexeme=new char[1];
 lexeme[0]=0;
 type=DEFAULT;
}
Token::Token(char *name)
{
 lexeme=new char[strlen(name)+1];
 strcpy(lexeme,name);
 type=DEFAULT;
}
Token::Token(char *name,int type)
{
 lexeme=new char[strlen(name)+1];
 strcpy(lexeme,name);
 Token::type=type;
}
Token::~Token()
{
 delete[] lexeme;
}

unsigned char Token::getOpcode(void)
{
 return opcode;
}
void Token::setOpcode(unsigned char code)
{
 opcode=code;
}
void Token::setName(char *name)
{
 delete[] lexeme;
 lexeme=new char[strlen(name)];
 strcpy(lexeme,name);
}
void Token::setType(int type)
{
 Token::type=type;
}
void Token::setLineNumber(int line)
{
 if(line<=0)
  fprintf(stderr,"Line number must be greater than 0\n");
 else
  linenum=line;
}
char * Token::getName(void)
{
 return lexeme;
}
int Token::getType(void)
{
 return type;
}
int Token::getLineNumber(void)
{
 return linenum;
}

class Operator
{
 char * name;
 unsigned char opcode;
 int type;
 public :

  Operator(char *op,unsigned char code,int optype);
  ~Operator();
  char * getName(void);
  int isEqual(char *op);
  unsigned char getOpcode(void);
  int getType(void);
};

Operator::Operator(char *op,unsigned char code,int optype)
{
 name=new char[strlen(op)+1];
 strcpy(name,op);
 opcode=code;
 type=optype;
}
Operator::~Operator()
{
 delete[] name;
}
char * Operator::getName(void)
{
 return name;
}
unsigned char Operator::getOpcode(void)
{
 return opcode;
}
int Operator::getType(void)
{
 return type;
}
int Operator::isEqual(char * op)
{
 return stricmp(op,name);
}

class OperatorTable
{
 Operator * table[31];
 int numofoperator;
 public :
  OperatorTable();
  ~OperatorTable();
  Operator * isOperator(char *name);
};
OperatorTable::OperatorTable()
{
  numofoperator=30;
  table[0]=new Operator("add",24,OPERATOR1);
  table[1]=new Operator("and",64,OPERATOR1);
  table[2]=new Operator("byte",1,BYTE);
  table[3]=new Operator("comp",40,OPERATOR1);
  table[4]=new Operator("div",36,OPERATOR1);
  table[5]=new Operator("end",1,END);
  table[6]=new Operator("j",60,OPERATOR2);
  table[7]=new Operator("jeq",48,OPERATOR1);
  table[8]=new Operator("jgt",52,OPERATOR1);
  table[9]=new Operator("jlt",56,OPERATOR1);
  table[10]=new Operator("jsub",72,OPERATOR1);
  table[11]=new Operator("lda",0,OPERATOR1);
  table[12]=new Operator("ldch",80,OPERATOR1);
  table[13]=new Operator("ldl",8,OPERATOR1);
  table[14]=new Operator("ldx",4,OPERATOR1);
  table[15]=new Operator("mul",32,OPERATOR1);
  table[16]=new Operator("or",68,OPERATOR1);
  table[17]=new Operator("rd",216,OPERATOR1);
  table[18]=new Operator("resb",1,RESB);
  table[19]=new Operator("resw",1,RESW);
  table[20]=new Operator("rsub",76,OPERATOR2);
  table[21]=new Operator("sta",12,OPERATOR1);
  table[22]=new Operator("start",1,START);
  table[23]=new Operator("stch",84,OPERATOR1);
  table[24]=new Operator("stl",20,OPERATOR1);
  table[25]=new Operator("stx",16,OPERATOR1);
  table[26]=new Operator("sub",28,OPERATOR1);
  table[27]=new Operator("td",224,OPERATOR1);
  table[28]=new Operator("tix",44,OPERATOR1);
  table[29]=new Operator("wd",220,OPERATOR1);
  table[30]=new Operator("word",1,WORD);
}

OperatorTable::~OperatorTable()
{
 int i;
 for(i=0;i<numofoperator;i++)
  delete table[i];
}

Operator * OperatorTable::isOperator(char *name)
{
 int upper,lower,current,cmp;
 upper=numofoperator;
 lower=0;

 while(upper>=lower)
 {
  current=lower+ (int)(upper-lower)/2;
  cmp=table[current]->isEqual(name);
  if(cmp==0)
   return table[current];
  else if(cmp>0)
	lower=current+1;
       else
	upper=current-1;
 }
 return NULL;
}

class SymbolEntry
{
 char *lexeme;
 int line;
 long int address;
 int doubledec;
 int type;
 SymbolEntry *next;
 public:
  SymbolEntry(char *,int ,long int,int);
  ~SymbolEntry();
  char * getName(void);
  int getLineNumber(void);
  long int getAddress(void);
  void setDoubleDec(void);
  int prevDecleration(void);
  int isEqual(char *);
  int getType(void);
  void setNext(SymbolEntry *);
  SymbolEntry * getNext(void);
};
SymbolEntry::SymbolEntry(char *name,int linenum,long int addr,int typeinfo)
{
 lexeme=new char[strlen(name)+1];
 strcpy(lexeme,name);
 line=linenum;
 address=addr;
 doubledec=0;
 type=typeinfo;
 next=NULL;
}
SymbolEntry::~SymbolEntry(void)
{
 delete[] lexeme;
}

char * SymbolEntry::getName(void)
{
 return lexeme;
}

int SymbolEntry::getType(void)
{
 return type;
}

void SymbolEntry::setNext(SymbolEntry * entry)
{
 next=entry;
}

SymbolEntry * SymbolEntry::getNext(void)
{
 return next;
}
int SymbolEntry::getLineNumber(void)
{
 return line;
}

long int SymbolEntry::getAddress(void)
{
 return address;
}

void SymbolEntry::setDoubleDec(void)
{
 doubledec=1;
}

int SymbolEntry::prevDecleration(void)
{
 return doubledec;
}

int SymbolEntry::isEqual(char * name)
{
 return stricmp(name,lexeme)?0:1;
}

class SymbolTable
{
 int numhashline;
 SymbolEntry * hashline[512];
 int hash(char *);
 public :
  SymbolTable();
  ~SymbolTable();
  SymbolEntry * inTable(char *);
  SymbolEntry * insertEntry(char *,int ,long int, int);
};
SymbolTable::SymbolTable(void)
{
 int i;
 numhashline=512;
 for(i=0;i<numhashline;i++)
  hashline[i]=NULL;
}

SymbolTable::~SymbolTable(void)
{
 int i;
 SymbolEntry *deleted;
 for(i=0;i<numhashline;i++)
 {
  while(hashline[i]!=NULL)
  {
   deleted=hashline[i];
   hashline[i]=deleted->getNext();
   delete deleted;
  }
 }
}

SymbolEntry * SymbolTable::inTable(char *name)
{
 int pos=hash(name);
 SymbolEntry *current=hashline[pos];

 while(current!=NULL)
 {
  if(current->isEqual(name))
   return current;
  else
   current=current->getNext();
 }
 return NULL;
}

SymbolEntry * SymbolTable::insertEntry(char *name,int line,long int address,int type)
{
 int pos;
 SymbolEntry *first,*second;
 if((first=inTable(name))==NULL)
 {
  pos=hash(name);
  first=hashline[pos];
  second=new SymbolEntry(name,line,address,type);
  second->setNext(first);
  hashline[pos]=second;
  return second;
 }
 else
 {
  first->setDoubleDec();
  fprintf(stderr,"Line %d: %s Declared previosly at line %d\n",line,first->getName(),first->getLineNumber());
  return NULL;
 }
}

int SymbolTable::hash(char *s)
{
 char *p,ch;
 unsigned h=0,g;

 for(p=s;*p!=0;p++)
 {
  if(*p>='A' && *p <='Z')
   ch=*p-'A'+'a';
  else
   ch=*p;
  h=(h<<4)+(ch);
  g=h&0xf0000000;
  if(g)
  {
   h=h^(g>>24);
   h=h^g;
  }
 }

 return h%numhashline;
}


class Lexer
{
 FILE * file;
 int line;
 int state;
 int EOFcond;
 char lexeme[100];
 char restofline[100];
 int lexemelength;
 char *fname;
 Token * tokenGenerate(int);
 OperatorTable *optable;
 char ch;
 void skipToNext(void);
 void skipToNext(int);
 public :
 int isDigit(void)
 {
  if(ch>='0' && ch <='9')
   return 1;
  else
   return 0;
 }
 int isHexDigit(void)
 {
  if(ch>='A' && ch<='F' || ch>='a' && ch<='f')
   return 1;
  else
   return 0;
 }

 int isLetter(void)
 {
  if(ch>='A' && ch<='Z' || ch>='a' && ch<='z')
   return 1;
  else
   return 0;
 }
 Lexer(char *,OperatorTable *);
 Token * next(void);
 void replay(void);
 char * skipLine(void);
};
Lexer::Lexer(char * filename,OperatorTable *optab)
{
 fname=new char[strlen(filename)+1];
 strcpy(fname,filename);
 file=fopen(filename,"r");
 if(file==NULL)
 {
  fprintf(stderr,"File could not be opened\n");
  exit(1);
 }
 optable=optab;
 line=1;
 state=0;
 if(fscanf(file,"%c",&ch)==EOF)
  EOFcond=1;
 else
  EOFcond=0;
}
void Lexer::replay(void)
{
 if(file!=NULL)
 {
  fclose(file);
  file=fopen(fname,"r");
  line=1;
  if(fscanf(file,"%c",&ch)==EOF)
   EOFcond=1;
  else
   EOFcond=0;
 }
}

void Lexer::skipToNext(void)
{
 while(!EOFcond && ch!=' ' && ch!='\t' && ch!=',' && ch!='\n')
 {
  if(fscanf(file,"%c",&ch)==EOF)
   EOFcond=1;
  else
  lexeme[lexemelength++]=ch;
 }
 lexeme[lexemelength-1]=0;
}

void Lexer::skipToNext(int arg)
{
 while(!EOFcond && ch!=' ' && ch!='\t' && ch!='\n')
 {
  if(fscanf(file,"%c",&ch)==EOF)
   EOFcond=1;
  else
  lexeme[lexemelength++]=ch;
 }
 lexeme[lexemelength-1]=0;
}


Token * Lexer::tokenGenerate(int tokentype)
{
 Operator *op;
 Token *token;
 if(tokentype==ERROR)
 {
  token=new Token("",ERROR);
  fprintf(stderr,"Line %d: ",line);
  if(state==8)
   fprintf(stderr,"Invalid Hexadecimal ");
  else if(state==10)
	fprintf(stderr,"Invalid Character String ");
       else
	fprintf(stderr,"Errenous token ");
  fprintf(stderr,"%s\n",lexeme);
  token->setName(lexeme);
 }
 else if(tokentype==EOFTOKEN)
       token=new Token("",EOFTOKEN);
      else if(tokentype==COMMA)
	    token=new Token("",COMMA);
	   else
	   {
	    if(tokentype==CHARACTER || tokentype==HEXADECIMAL)
	    {
	     if(!EOFcond && ch!=' ' && ch!='\n' && ch!='\t')
	     {
	      skipToNext(1);
	      tokentype=ERROR;
	     }
	    }
	    else
	    {
	     if(!EOFcond && ch!=' ' && ch!='\n' && ch!='\t' && ch!=',')
	     {
	      skipToNext();
	      tokentype=ERROR;
	     }
	    }

	    lexeme[lexemelength-1]=0;
	    token=new Token(lexeme,tokentype);
	    if(tokentype!=ERROR && (op=optable->isOperator(lexeme))!=NULL)
	    {
	     token->setType(op->getType());
	     token->setOpcode(op->getOpcode());
	    }
	   }
 token->setLineNumber(line);
 return token;
}
char * Lexer::skipLine()
{
 int i=0;
 while(ch!='\n' && !EOFcond && fscanf(file,"%c",&ch)!=EOF)
  restofline[i++]=ch;
 restofline[i]=0;
 if(ch!='\n')
  EOFcond=1;
 return restofline;
}

Token * Lexer::next(void)
{
 int intret;

 if(EOFcond)
  return new Token("",EOFTOKEN);
 else
 {
  state=0;
  lexeme[0]=ch;
  lexemelength=1;
  while(!EOFcond)
  {
   switch(state)
   {
    case 0:{
	    if(ch=='X' || ch=='x')
	     state=1;
	    else if(ch=='C' || ch=='c')
		  state=2;
		 else if(ch=='.')
		       state=3;
		      else if(ch=='+' || ch=='-')
			    state=4;
			   else if(isDigit())
				{
				 intret=1;
				 state=5;
				}
				else if(isLetter())
				      state=6;
				     else if(ch==' ' || ch=='\t' || (ch=='\n'?++line:0))
					   state=7;
					  else if(ch==',')
						state=15;
					       else
					       {
						skipToNext();
						return tokenGenerate(ERROR);
					       }
	    break;
	   }
    case 1:{
	    if(ch=='\'')
	     state=8;
	    else
	     state=61;
	    break;
	   }
    case 2:
	   {
	    if(ch=='\'')
	     state=10;
	    else
	     state=61;
	    break;
	   }
    case 3:
	   {
	    if(ch=='\n')
	     state=0;
	    else
	     state=3;
	    break;
	   }
    case 4:
	   {
	    if(isDigit())
	     state=5;
	    else
	    {
	     skipToNext();
	     return tokenGenerate(ERROR);
	    }
	    break;
	   }
    case 5:
	   {
	    if(isDigit())
	     state=5;
	    else if(isHexDigit())
		 {
		  intret=0;
		  state=5;
		 }
		 else
		  state=13;
	    break;
	   }
    case 6:{
	    if(isLetter() || isDigit())
	     state=6;
	    else
	     state=14;
	    break;
	   }
    case 7:
	   {
	    if(ch==' ' || ch=='\t' || (ch=='\n'?++line:0))
	     state=7;
	    else
	    {
	     state=0;
	     lexeme[0]=ch;
	     lexemelength=1;
	    }
	    break;
	   }
    case 8:
	   {
	    if(isHexDigit() || isDigit())
	     state=8;
	    else if(ch=='\'')
		  state=9;
		 else
		 {
		  skipToNext();
		  return tokenGenerate(ERROR);
		 }
	    break;
	   }
    case 9:
	   {
	    return tokenGenerate(HEXADECIMAL);
	   }
    case 10:
	   {
	    if(ch=='\'')
	     state=11;
	    else if(ch=='\n')
		  return tokenGenerate(ERROR);
	    break;
	   }
    case 11:
	    {
	     return tokenGenerate(CHARACTER);
	    }
    case 13:
	    {
	     return intret?tokenGenerate(INTEGER):tokenGenerate(HEXINT);
	    }
    case 14:
	    {
	     if(EOFcond || ch==' ' || ch=='\n' || ch=='\t' || ch==',')
	      return tokenGenerate(LABEL);
	     else
	     {
	      skipToNext();
	      return tokenGenerate(ERROR);
	     }
	    }
    case 15:
	    {
	     return tokenGenerate(COMMA);
	    }

   }
   if(state==61)
    state=6;
   else if(state && state!=13 && state!=14)
	{
	 if(fscanf(file,"%c",&ch)==EOF)
	 {
	  ch=EOFCHAR;
	  EOFcond=1;
	 }
	 else if(state!=7)
	      {
	       lexeme[lexemelength++]=ch;
	      }
	}
  }
  return tokenGenerate(EOFTOKEN);
 }
}

class FileManager
{
 FILE * file;
 char *fname;
 public:
  FileManager(char *);
 ~FileManager();
  void fileClose(void);
  void fileDestroy(void);
  void writeToFile(char *);
};
FileManager::FileManager(char *f)
{
 fname=(char *)malloc(strlen(f)*sizeof(char));
 strcpy(fname,f);
 file=fopen(fname,"w");
}

FileManager::~FileManager(void)
{
 if(file!=NULL)
 {
  fclose(file);
  file=NULL;
 }
}
void FileManager::fileClose(void)
{
 if(file!=NULL)
 {
  fclose(file);
  file=NULL;
 }
}
void FileManager::fileDestroy(void)
{
 if(file!=NULL)
 {
  fclose(file);
  file=fopen(fname,"w");
  if(file!=NULL)
  {
   fclose(file);
   file=NULL;
  }
 }
}

void FileManager::writeToFile(char *str)
{
 if(file!=NULL)
 {
  fprintf(file,"%s",str);
 }
}


class Parser
{
 SymbolEntry *startlabel;
 SymbolTable * symtab;
 FileManager *fmngr,*omngr;
 char adrstring[7];
 char opstr[7];
 int shiftconds[16];
 int pass1active;
 int pass1ok;
 long int adrescntr,length;
 Lexer *lex;
 char fline[100];
 char record[71];
 int last;
 long int addresscntr;
 long int hextoint(char *);
 long int strtoint(char *);
 char * opcode(unsigned char,long int);
 char * inttohex(int,long int);
 void append(char *);
 void appendchars(char *);

 void closerecord(void);
 public :
  Parser(Lexer *,SymbolTable *,FileManager *,FileManager *);
  void pass1(void);
  void pass2(void);
};

Parser::Parser(Lexer * lexer,SymbolTable *table,FileManager *fm,FileManager *om)
{
 lex=lexer;
 symtab=table;
 fmngr=fm;
 omngr=om;
 length=0;
 addresscntr=0;
 pass1ok=1;
 pass1active=0;
 last=0;
}

void Parser::closerecord(void)
{
 record[last++]='\n';
 record[last]=0;
 last=0;
}

void Parser::appendchars(char *str)
{
 int i,j,k,m;
 for(i=0,j=strlen(str);i<j;i++)
 {
  k=str[i];
  for(m=1;m>=0;m--)
  {
   if(k%16>10)
    record[last+m]='A'+k%16-10;
   else
    record[last+m]='0'+k%16;
   k=k/16;
  }
  last+=2;
 }
}
void Parser::append(char *str)
{
 int i,j=strlen(str);

 for(i=0;i<j;i++)
 {
  record[last++]=str[i];
 }
}
char * Parser::opcode(unsigned char code,long int adr)
{
  int i,j;
  opstr[6]=0;
  for(i=1;i>=0;i--)
  {
   j=code%16;
   if(j>9)
    opstr[i]='A'+j-10;
   else
    opstr[i]='0'+j;
   code/=16;
  }
  for(i=5;i>=2;i--)
  {
   j=adr%16;
   if(j>9)
    opstr[i]='A'+j-10;
   else
    opstr[i]='0'+j;
   adr/=16;
  }
  return opstr;
}

char * Parser::inttohex(int size,long int num)
{
 int i,j;
 adrstring[size]=0;
 for(i=size-1;i>=0;i--)
 {
  j=num%16;
  if(j>9)
   adrstring[i]='A'+j-10;
  else
   adrstring[i]='0'+j;
  num/=16;
 }
 return adrstring;
}


long int Parser::hextoint(char *adr)
{
 long int result=0,mul=1,tmp=0,jm=1;
 int j=0,i=strlen(adr);

 if(adr[0]=='+')
 {
  j=1;
 }
 else if(adr[0]=='-')
      {
       j=1;
       jm=-1;
      }

 while(--i>=j)
 {
  if(adr[i]>='A' && adr[i]<='Z')
   tmp=adr[i]-'A'+10;
  else if(adr[i]>='a' && adr[i]<='z')
	tmp=adr[i]-'a'+10;
       else
	tmp=adr[i]-'0';
  result+=tmp*mul;
  mul*=16;
 }
 return jm*result;
}

long int Parser::strtoint(char *str)
{
 long int result=0,mul=1,jm=1;
 int j=0,i=strlen(str);

 if(str[0]=='+')
 {
  j=1;
  jm=1;
 }
 else if(str[0]=='-')
      {
       j=1;
       jm=-1;
      }

 while(--i>=j)
 {
  result+=mul*(str[i]-'0');
  mul*=10;
 }
 return jm*result;
}

void Parser::pass1(void)
{
 Token *token,*stoken,*optoken;
 SymbolEntry *entry;
 int type,state,enable,notok,errorcntr,cmline,prev;

 pass1active=1;
 errorcntr=addresscntr=length=0;
 token=stoken=optoken=NULL;
 entry=NULL;
 stoken=lex->next();
 if(stoken->getType()!=LABEL)
 {
  errorcntr++;
  pass1ok=0;
  fprintf(stderr,"Line %d: Program name expected but %s found\n",stoken->getLineNumber(),stoken->getName());
  if(stoken->getType()==START)
   optoken=stoken;
  else
   optoken=lex->next();
 }
 else
 {
  optoken=lex->next();
 }

 if(optoken->getType()==START)
 {
  token=lex->next();
  type=token->getType();
  if(type==HEXINT || type==INTEGER)
  {
   addresscntr=hextoint(token->getName());
   if(addresscntr<0)
   {
    fprintf(stderr,"Line %d: Address cannot be negative\n",token->getLineNumber());
    errorcntr++;
    pass1ok=0;
   }
   else if(addresscntr>(long int) 0x0fff)
	{
	 fprintf(stderr,"Line %d: Address exceeds the memory size\n",token->getLineNumber());
	 errorcntr++;
	 pass1ok=0;
	}
	else
	{
	 startlabel=symtab->insertEntry(stoken->getName(),stoken->getLineNumber(),addresscntr+length,LABEL);
	 lex->skipLine();
	}
   token=lex->next();
  }
  else
  {
   errorcntr++;
   pass1ok=0;
   fprintf(stderr,"Line %d: Address expected\n",token->getLineNumber());
   if(token->getLineNumber()==optoken->getLineNumber())
   {
    lex->skipLine();
    token=lex->next();
   }
  }
 }
 else
  token=optoken;

 type=token->getType();
 while(type!=END)
 {
  state=0;
  notok=1;
  while(notok)
  {
   if(type==EOFTOKEN)
   {
    fprintf(stderr,"Line %d: Unexpected end of file\n",token->getLineNumber());
    exit(1);
   }
   if(errorcntr>10)
   {
    fprintf(stderr,"Fatal error: Confused for too many errors\n");
    exit(1);
   }
   enable=1;
   switch(state)
   {
    case 0:{
	    switch(type)
	    {
	     case LABEL:{
			  state=1;
			  if(symtab->insertEntry(token->getName(),token->getLineNumber(),addresscntr+length,LABEL)==NULL)
			  {
			   errorcntr++;
			   pass1ok=0;
			  }
			  break;
			 }
	     case OPERATOR1:{
			     optoken=token;
			     state=2;
			     break;
			    }
	     case OPERATOR2:{
			     length+=3;
			     lex->skipLine();
			     notok=0;
			     break;
			    }
	     case END      :{
			     notok=0;
			     enable=0;
			     break;
			    }
	     default       :{
			     errorcntr++;
			     pass1ok=0;
			     delete token;
			     token=lex->next();
			     type=token->getType();
			     enable=0;
			     if(type!=OPERATOR1 && type!=OPERATOR2 && type!=RESW && type!=RESB && type!=WORD && type!=BYTE)
			     {
			      fprintf(stderr,"Line %d: Invalid operator %s\n",token->getLineNumber(),token->getName());
			     }
			     else
			     {
			      fprintf(stderr,"Line %d: Invalid label %s\n",token->getLineNumber(),token->getName());
			      state=1;
			     }
			     break;
			    }
	    }
	    break;
	   }
    case 1:{
	    switch(type)
	    {
	     case OPERATOR1:{
			     state=2;
			     optoken=token;
			     break;
			    }
	     case OPERATOR2:{
			     length+=3;
			     lex->skipLine();
			     notok=0;
			     break;
			    }
	     case RESW      :{
			      optoken=token;
			      state=6;
			      break;
			     }
	     case RESB      :{
			      optoken=token;
			      state=7;
			      break;
			     }
	     case WORD      :{
			      optoken=token;
			      state=8;
			      break;
			     }
	     case BYTE      :{
			      optoken=token;
			      state=9;
			      break;
			     }
	     default        :{
			      pass1ok=0;
			      errorcntr++;
			      enable=0;
			      delete token;
			      token=lex->next();
			      type=token->getType();
			      fprintf(stderr,"Line %d: ",token->getLineNumber());
			      if(type==CHARACTER || type==HEXADECIMAL || type==INTEGER)
			      {
			       fprintf(stderr,"Constant decleration keyword expected but %s found",token->getName());
			       state=9;
			      }
			      else if(type==HEXINT || LABEL)
				   {
				    fprintf(stderr,"Operator expected but %s found\n",token->getName());
				    state=2;
				   }
				   else
				   {
				    fprintf(stderr,"Unresolved error aroused by %s\n",token->getName());
				    state=2;
				   }
			      break;
			     }
	    }
	    break;
	   }
    case 2:{
	    switch(type)
	    {
	     case LABEL :{
			  prev=token->getLineNumber();
			  state=3;
			  break;
			 }
	     case HEXINT :{
			   prev=token->getLineNumber();
			   state=3;
			   if(hextoint(token->getName())>0xfff)
			   {
			    fprintf(stderr,"Line %d: %s exceeds memory space\n",token->getLineNumber(),token->getName());
			    errorcntr++;
			    pass1ok=0;
			   }
			   break;
			  }
	     default     :{
			   pass1ok=0;
			   errorcntr++;
			   fprintf(stderr,"Line %d: Operand expected but %s found\n",token->getLineNumber(),token->getName());
			   state=3;
			   break;
			  }
	    }
	    break;
	   }
    case 3:{
	    if(type==COMMA)
	    {
	     cmline=token->getLineNumber();
	     state=4;
	    }
	    else
	    {
	     length+=3;
	     notok=0;
	     if(prev==token->getLineNumber())
	     {
	      lex->skipLine();
	     }
	     else
	      enable=0;
	    }
	    break;
	   }
    case 4:{
	    notok=0;
	    if(type==LABEL)
	    {
	     if(stricmp(token->getName(),"x")==0)
	     {
	      length+=3;
	      lex->skipLine();
	     }
	     else
	     {
	      errorcntr++;
	      pass1ok=0;
	      fprintf(stderr,"Line %d: Register X expected\n",token->getLineNumber());
	      if(cmline==token->getLineNumber())
	      {
	       lex->skipLine();
	      }
	      else
	       enable=0;
	     }
	    }
	    else
	    {
	     errorcntr++;
	     pass1ok=0;
	     fprintf(stderr,"Line %d: Register X expected\n",token->getLineNumber());
	     if(cmline==token->getLineNumber())
	     {
	      lex->skipLine();
	     }
	     else
	      enable=0;
	    }
	    break;
	   }
   case 6 :{
	    if(type==INTEGER)
	    {
	     length+=3*strtoint(token->getName());
	     notok=0;
	     lex->skipLine();
	    }
	    else
	    {
	     errorcntr++;
	     pass1ok=0;
	     fprintf(stderr,"Line %d: Integer expected\n",token->getLineNumber());
	     if(token->getLineNumber()==optoken->getLineNumber())
	     {
	      lex->skipLine();
	     }
	     else
	      enable=0;
	     notok=0;
	    }
	    break;
	   }
    case 7:{
	    if(type==INTEGER)
	    {
	     length+=strtoint(token->getName());
	     notok=0;
	     lex->skipLine();
	    }
	    else
	    {
	     errorcntr++;
	     pass1ok=0;
	     fprintf(stderr,"Line %d: Integer expected\n",token->getLineNumber());
	     if(token->getLineNumber()==optoken->getLineNumber())
	     {
	      lex->skipLine();
	     }
	     else
	      enable=0;
	     notok=0;
	    }
	    break;
	   }

    case 8:{
	    switch(type)
	    {
	     case HEXADECIMAL:{
			       if((strlen(token->getName())-3)%6==0)
			       {
				length+=(strlen(token->getName())-3)/2;
			       }
			       else
			       {
				errorcntr++;
				fprintf(stderr,"Line %d: The hexadecimal %s does not form a word\n",token->getLineNumber(),token->getLineNumber());
				fprintf(stderr,"   *   : 6 hex digit forms a word\n");
			       }
			       notok=0;
			       lex->skipLine();
			       break;
			      }
	     case CHARACTER  :{
			       if((strlen(token->getName())-3)%3==0)
			       {
				length+=strlen(token->getName())-3;
			       }
			       else
			       {
				errorcntr++;
				fprintf(stderr,"Line %d: The character %s does not form a word\n",token->getLineNumber(),token->getName());
				fprintf(stderr,"   *   : 3 character forms a word\n");
			       }
			       notok=0;
			       lex->skipLine();
			       break;
			      }
	     case INTEGER    :{
			       length+=3;
			       notok=0;
			       lex->skipLine();
			       break;
			      }
	     default         :{
			       pass1ok=0;
			       errorcntr++;
			       fprintf(stderr,"Line %d: A constant expected but %s found\n",token->getLineNumber(),token->getName());
			       if(token->getLineNumber()==optoken->getLineNumber())
			       {
				lex->skipLine();
			       }
			       else
				enable=0;
			       notok=0;
			       break;
			      }
	    }
	    break;
	   }
    case 9:{
	    switch(type)
	    {
	     case INTEGER    :{
			       length++;
			       notok=0;
			       lex->skipLine();
			       if(strtoint(token->getName())>255)
				fprintf(stderr,"Warning Line %d: Constant out of range (casting occurs)\n",token->getLineNumber());
			       break;
			      }
	     case HEXADECIMAL:{
			       if((strlen(token->getName())-3)%2==0)
			       {
				lex->skipLine();
				length+=(strlen(token->getName())-3)/2;
			       }
			       else
			       {
				errorcntr++;
				fprintf(stderr,"Line %d: The hexadecimal %s does not form a byte\n",token->getLineNumber(),token->getName());
				fprintf(stderr,"   *   : 2 hexadecimal digit form a byte\n");
			       }
			       notok=0;
			       break;
			      }
	     case CHARACTER  :{
			       notok=0;
			       lex->skipLine();
			       length+=strlen(token->getName())-3;
			       break;
			      }
	     default         :{
			       pass1ok=0;
			       errorcntr++;
			       fprintf(stderr,"Line %d: A constant expected but %s found\n",token->getLineNumber(),token->getName());
			       if(token->getLineNumber()==optoken->getLineNumber())
			       {
				lex->skipLine();
			       }
			       else
				enable=0;
			       notok=0;
			       break;
			      }

	    }
	    break;
	   }
   }

   if(enable)
   {
    delete token;
    token=lex->next();
    type=token->getType();
   }
  }
 }
 token=lex->next();
 type=token->getType();
 if(type==LABEL)
 {
  if((entry=symtab->inTable(token->getName()))==NULL)
  {
   fprintf(stderr,"Line %d: There is no address named by %s\n",token->getLineNumber(),token->getName());
   pass1ok=0;
  }
  else
  {
    startlabel=entry;
  }
 }
 else if(type!=EOFTOKEN)
      {
       pass1ok=0;
       fprintf(stderr," !  Garbage after program END\n");
       fprintf(stderr,"Line %d : Remove them\n",token->getLineNumber());
      }
}


void Parser::pass2(void)
{
 Token *adrtoken,*token,*stoken,*ltoken,*optoken,*opndtoken;
 int type,state,enable,notok,errorcntr;
 unsigned int i;
 long int address,codecntr;
 char *name,ch;
 SymbolEntry * operand;

 if(pass1active==0)
 {
  fprintf(stderr," Pass1 has not been called\n First call it\n");
  exit(1);
 }
 else if(pass1ok==0)
      {
       fprintf(stderr,"Pass1 : Grammatical or lexical errors\n   Please correct the errors\n");
       exit(1);
      }

 if(addresscntr+length>0xfff)
 {
  fprintf(stderr,"   *  : Program does not fit into address space\n");
  fprintf(stderr,"   !  : Start address + program size exceeds 32k space\n");
  exit(1);
 }

 errorcntr=0;
 token=stoken=ltoken=optoken=opndtoken=NULL;
 lex->replay();
 stoken=lex->next();
 name=stoken->getName();
 optoken=lex->next();
 if(optoken->getType()==START)
 {
  adrtoken=lex->next();
  lex->skipLine();
  token=lex->next();
 }
 else
 {
  token=optoken;
  optoken=NULL;
 }

 append("H");
 if(strlen(name)>=6)
 {
  name[6]=0;
  append(name);
 }
 else
 {
  for(i=0;i<strlen(name);i++)
   opstr[i]=name[i];
  for(;i<6;i++)
   opstr[i]=' ';
  opstr[6]=0;
  append(opstr);
 }

  if(optoken!=NULL)
  {
   sprintf(fline,"%s\t\t%s\t%s\t%s\n",inttohex(4,addresscntr),stoken->getName(),optoken->getName(),adrtoken->getName());
   append(inttohex(6,addresscntr));
   append(inttohex(6,length));
  }
  else
  {
   sprintf(fline,"%s\t\t%s\n",inttohex(4,addresscntr),stoken->getName());
  }

 closerecord();
 omngr->writeToFile(record);
 fmngr->writeToFile(fline);
 length=0; // obj file length reset
 type=token->getType();
 codecntr=0;
 last=0;

 while(type!=END)
 {
  state=0;
  notok=1;
  while(notok)
  {
   enable=1;
   switch(state)
   {
    case 0:{
	    switch(type)
	    {
	     case LABEL:{
			  state=1;
			  ltoken=token;
			  break;
			 }
	     case OPERATOR1:{
			     ltoken=new Token("",LABEL);
			     optoken=token;
			     state=2;
			     break;
			    }
	     case OPERATOR2:{
			     sprintf(fline,"%s\t\t\t%s\t\t\t%s\n",inttohex(4,addresscntr+length),token->getName(),(name=opcode(token->getOpcode(),0)));
			     codecntr+=3;
			     if(codecntr>30)
			     {
			      closerecord();
			      last=7;
			      append(inttohex(2,codecntr-3));
			      codecntr=3;
			      last=0;
			      omngr->writeToFile(record);
			     }
			     if(last==0)
			     {
			      append("T");
			      append(inttohex(6,addresscntr+length));
			      append("  ");
			     }
			     append(name);
			     length+=3;
			     fmngr->writeToFile(fline);
			     delete token;
			     lex->skipLine();
			     notok=0;
			     break;
			    }
	     case END      :{
			     notok=0;
			     enable=0;
			     break;
			    }
	    }
	    break;
	   }
    case 1:{
	    switch(type)
	    {
	     case OPERATOR1:{
			     state=2;
			     optoken=token;
			     break;
			    }
	     case OPERATOR2:{
			     sprintf("%s\t\t%s\t%s\t\t%s\n",inttohex(4,addresscntr+length),ltoken->getName(),token->getName(),(name=opcode(token->getOpcode(),0)));
			     fmngr->writeToFile(fline);

			     codecntr+=3;
			     if(codecntr>30)
			     {
			      closerecord();
			      last=7;
			      append(inttohex(2,codecntr-3));
			      codecntr=3;
			      last=0;
			      omngr->writeToFile(record);
			     }
			     length+=3;
			     if(last==0)
			     {
			      append("T");
			      append(inttohex(6,addresscntr+length));
			      append("  ");
			     }
			     append(name);
			     lex->skipLine();
			     delete token;
			     delete ltoken;
			     notok=0;
			     break;
			    }
	     case RESW      :{
			      optoken=token;
			      state=6;
			      break;
			     }
	     case RESB      :{
			      optoken=token;
			      state=7;
			      break;
			     }
	     case WORD      :{
			      optoken=token;
			      state=8;
			      break;
			     }
	     case BYTE      :{
			      optoken=token;
			      state=9;
			      break;
			     }
	    }
	    break;
	   }
    case 2:{
	    state=3;
	    if(token->getType()==LABEL)
	    {
	     if((operand=symtab->inTable(token->getName()))==NULL)
	     {
	      fprintf(stderr,"Line %d: Undefined label %s\n",token->getLineNumber(),token->getName());
	      errorcntr++;
	     }
	     else
	     {
	      address=operand->getAddress();
	     }
	    }
	    else
	     address=hextoint(token->getName());

	    opndtoken=token;
	    break;
	   }
    case 3:{
	    if(type==COMMA)
	    {
	     state=4;
	     address+=32768;
	    }
	    else
	    {
	     sprintf(fline,"%s\t\t%s\t%s\t%s\t\t%s\n",inttohex(4,addresscntr+length),ltoken->getName(),optoken->getName(),opndtoken->getName(),(name=opcode(optoken->getOpcode(),address)));
	     fmngr->writeToFile(fline);
	     codecntr+=3;
	     if(codecntr>30)
	     {
	      closerecord();
	      last=7;
	      append(inttohex(2,codecntr-3));
	      codecntr=3;
	      last=0;
	      omngr->writeToFile(record);
	     }
	     if(last==0)
	     {
	      append("T");
	      append(inttohex(6,addresscntr+length));
	      append("  ");
	     }
	     append(name);
	     length+=3;
	     notok=0;
	     if(optoken->getLineNumber()==token->getLineNumber())
	     {
	      lex->skipLine();
	      delete ltoken;
	      delete optoken;
	      delete opndtoken;
	      delete token;
	     }
	     else
	      enable=0;
	    }
	    break;
	   }
    case 4:{
	    notok=0;
	    sprintf(fline,"%s\t\t%s\t%s\t%s,%s\t\t%s\n",inttohex(4,addresscntr+length),ltoken->getName(),optoken->getName(),opndtoken->getName(),token->getName(),(name=opcode(optoken->getOpcode(),address)));
	    fmngr->writeToFile(fline);
	    codecntr+=3;
	    if(codecntr>30)
	    {
	     closerecord();
	     last=7;
	     append(inttohex(2,codecntr-3));
	     codecntr=3;
	     last=0;
	     omngr->writeToFile(record);
	    }
	    if(last==0)
	    {
	     append("T");
	     append(inttohex(6,addresscntr+length));
	     append("  ");
	    }
	    append(name);
	    length+=3;
	    delete ltoken;
	    delete optoken;
	    delete opndtoken;
	    delete token;

	    lex->skipLine();
	    break;
	   }
   case 6 :{
	    sprintf(fline,"%s\t\t%s\t%s\t%s\n",inttohex(4,addresscntr+length),ltoken->getName(),optoken->getName(),token->getName());
	    fmngr->writeToFile(fline);
	    if(last>8)
	    {
	     closerecord();
	     last=7;
	     append(inttohex(2,codecntr));
	     last=0;
	     codecntr=0;
	     omngr->writeToFile(record);
	    }
	    length+=3*strtoint(token->getName());
	    notok=0;
	    delete ltoken;
	    delete optoken;
	    delete token;
	    lex->skipLine();
	    break;
	   }
    case 7:{
	    sprintf(fline,"%s\t\t%s\t%s\t%s\n",inttohex(4,addresscntr+length),ltoken->getName(),optoken->getName(),token->getName());
	    fmngr->writeToFile(fline);
	    if(last>8)
	    {
	     closerecord();
	     last=7;
	     append(inttohex(2,codecntr));
	     last=0;
	     codecntr=0;
	     omngr->writeToFile(record);
	    }
	    length+=strtoint(token->getName());
	    notok=0;
	    delete ltoken;
	    delete optoken;
	    delete token;
	    lex->skipLine();
	    break;
	   }

    case 8:{
	    switch(type)
	    {
	     case HEXADECIMAL:{
			       sprintf(fline,"%s\t\t%s\t%s\t%s\n",inttohex(4,addresscntr+length),ltoken->getName(),optoken->getName(),token->getName());
			       fmngr->writeToFile(fline);
			       name=token->getName();
			       name[strlen(name)-1]=0;
			       name+=2;
			       while(strlen(name)>0)
			       {
				if(last==0)
				{
				 append("T");
				 append(inttohex(6,addresscntr+length));
				 append("  ");
				}
				if(strlen(name)/2+codecntr>30)
				{
				 i=(60-2*codecntr)/3*3;
				 ch=name[i];
				 name[i]=0;
				 length+=strlen(name)/2;
				 append(name);
				 closerecord();
				 last=7;
				 append(inttohex(2,codecntr+i));
				 last=0;
				 name[i]=ch;
				 name+=i;
				 omngr->writeToFile(record);
				 codecntr=0;
				}
				else
				{
				 append(name);
				 codecntr+=strlen(name)/2;
				 length+=strlen(name)/2;
				 name[0]=0;
				}
			       }
			       delete ltoken;
			       delete optoken;
			       delete token;
			       notok=0;
			       lex->skipLine();
			       break;
			      }
	     case CHARACTER  :{
			       sprintf(fline,"%s\t\t%s\t%s\t%s\n",inttohex(4,addresscntr+length),ltoken->getName(),optoken->getName(),token->getName());
			       fmngr->writeToFile(fline);
			       name=token->getName();
			       name[strlen(name)-1]=0;
			       name+=2;
			       while(strlen(name)>0)
			       {
				if(strlen(name)+codecntr>30)
				{
				 i=30-codecntr;
				 ch=name[i];
				 name[i]=0;
				 length+=strlen(name);
				 appendchars(name);
				 closerecord();
				 last=7;
				 append(inttohex(2,30));
				 name[i]=ch;
				 name+=i;
				 omngr->writeToFile(record);
				 codecntr=0;
				}
				else
				{
				 appendchars(name);
				 codecntr+=strlen(name);
				 length+=strlen(name);
				 name[0]=0;
				}
			       }
			       delete ltoken;
			       delete optoken;
			       delete token;
			       notok=0;
			       lex->skipLine();
			       break;
			      }
	     case INTEGER    :{
			       sprintf(fline,"%s\t\t%s\t%s\t%s\n",inttohex(4,addresscntr+length),ltoken->getName(),optoken->getName(),token->getName());
			       fmngr->writeToFile(fline);
			       name=token->getName();
			       codecntr+=3;
			       if(codecntr>30)
			       {
				closerecord();
				last=7;
				append(inttohex(2,codecntr-3));
				codecntr=3;
				last=0;
				omngr->writeToFile(record);
			       }
			       if(last==0)
			       {
				append("T");
				append(inttohex(6,addresscntr+length));
				append("  ");
			       }
			       append(inttohex(6,strtoint(name)));
			       length+=3;
			       notok=0;
			       delete ltoken;
			       delete token;
			       lex->skipLine();
			       break;
			      }
	    }
	    break;
	   }
    case 9:{
	    switch(type)
	    {
	     case INTEGER    :{
			       sprintf(fline,"%s\t\t%s\t%s\t%s\n",inttohex(4,addresscntr+length),ltoken->getName(),optoken->getName(),token->getName());
			       fmngr->writeToFile(fline);
			       name=token->getName();
			       if(codecntr==30)
			       {
				closerecord();
				last=7;
				append(inttohex(2,codecntr-3));
				codecntr=0;
				last=0;
				omngr->writeToFile(record);
			       }
			       if(last==0)
			       {
				append("T");
				append(inttohex(6,addresscntr+length));
				append("  ");
			       }
			       append(inttohex(2,strtoint(name)));

			       length++;
			       codecntr++;
			       notok=0;
			       delete ltoken;
			       delete optoken;
			       delete token;
			       lex->skipLine();
			       break;
			      }
	     case HEXADECIMAL:{
			       sprintf(fline,"%s\t\t%s\t%s\t%s\n",inttohex(4,addresscntr+length),ltoken->getName(),optoken->getName(),token->getName());
			       fmngr->writeToFile(fline);
			       name=token->getName();
			       name[strlen(name)-1]=0;
			       name+=2;
			       while(strlen(name)>0)
			       {
				if(last==0)
				{
				 append("T");
				 append(inttohex(6,addresscntr+length));
				 append("  ");
				}
				if(strlen(name)/2+codecntr>30)
				{
				 i=(60-2*codecntr)/3*3;
				 ch=name[i];
				 name[i]=0;
				 length+=strlen(name)/2;
				 append(name);
				 closerecord();
				 last=7;
				 append(inttohex(2,codecntr+i));
				 last=0;
				 name[i]=ch;
				 name+=i;
				 omngr->writeToFile(record);
				 codecntr=0;
				}
				else
				{
				 append(name);
				 codecntr+=strlen(name)/2;
				 length+=strlen(name)/2;
				 name[0]=0;
				}


			       }
			       lex->skipLine();
			       notok=0;
			       delete ltoken;
			       delete optoken;
			       delete token;
			       break;
			      }
	     case CHARACTER  :{
			       sprintf(fline,"%s\t\t%s\t%s\t%s\n",inttohex(4,addresscntr+length),ltoken->getName(),optoken->getName(),token->getName());
			       fmngr->writeToFile(fline);
			       name=token->getName();
			       name[strlen(name)-1]=0;
			       name+=2;
			       while(strlen(name)>0)
			       {
				if(strlen(name)+codecntr>30)
				{
				 i=30-codecntr;
				 ch=name[i];
				 name[i]=0;
				 length+=strlen(name);
				 appendchars(name);
				 closerecord();
				 last=7;
				 append(inttohex(2,30));
				 last=0;
				 name[i]=ch;
				 name+=i;
				 omngr->writeToFile(record);
				 codecntr=0;
				}
				else
				{
				 appendchars(name);
				 codecntr+=strlen(name);
				 length+=strlen(name);
				 name[0]=0;
				}
			       }
			       delete ltoken;
			       delete optoken;
			       delete token;

			       notok=0;
			       lex->skipLine();
			       break;
			      }

	    }
	    break;
	   }
   }

   if(enable)
   {
    token=lex->next();
    type=token->getType();
   }
  }
 }
 stoken=token;
 token=lex->next();
 if(token->getType()==LABEL)
 {
  sprintf(fline,"\t\t%s\t%s\n",stoken->getName(),token->getName());
  if((operand=symtab->inTable(token->getName()))==NULL)
  {
   errorcntr++;
   fprintf(stderr,"Line %d: No label found with name %s\n",token->getLineNumber(),token->getName());
  }
  else
  {
   addresscntr=operand->getAddress();
  }
 }
 else
  sprintf(fline,"\t\t%s",stoken->getName());

 if(codecntr!=0)
 {
  closerecord();
  last=7;
  append(inttohex(2,codecntr));
  omngr->writeToFile(record);
  last=0;
 }
 append("E");
 append(inttohex(6,addresscntr));
 closerecord();
 omngr->writeToFile(record);

 fmngr->writeToFile(fline);
 delete stoken;
 delete token;
 if(errorcntr>0)
 {
  fprintf(stderr,"Pass2 : Errors occured\n");
  fprintf(stderr,"  !   : Please correct them\n");
  fmngr->fileDestroy();
  exit(1);
 }

 fmngr->fileClose();
}

class Assembler
{
 OperatorTable * optab;
 Lexer *lex;
 FileManager *f;
 FileManager *o;
 SymbolTable *symtab;
 Parser *parser;
 public:
  Assembler(char *);
  ~Assembler();
  void assemble(void);
};

Assembler::Assembler(char *fname)
{
  optab=new OperatorTable();
  lex=new Lexer(fname,optab);
  symtab=new SymbolTable();
  f=new FileManager("LISTFILE");
  o=new FileManager("OBJFILE");
  parser=new Parser(lex,symtab,f,o);
}
Assembler::~Assembler()
{
 delete optab;
 delete lex;
 delete symtab;
 delete f;
 delete o;
 delete parser;
}

void Assembler::assemble(void)
{
 parser->pass1();
 parser->pass2();
}

int main(int argc,char * argv[])
{
 Assembler  * asmb;

 if(argc==2)
 {
  asmb=new Assembler(argv[1]);
  asmb->assemble();
  delete asmb;
 }
  else
  {
   fprintf(stderr," Please specify the name of the source file\n");
  }

 return 0;
}
